import 'dart:async';

import 'package:collection/collection.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutterx_application/src/application.dart';
import 'package:flutterx_application/src/ui/fullscreen_draggable_bottom_sheet.dart';
import 'package:flutterx_application/src/utils/code.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'route/route_activity.dart';
part 'route/route_bottom_sheet.dart';
part 'route/route_dialog.dart';

///Arguments that can be passed to a route
typedef RouteArgs = Map<String, dynamic>;

/// Builder of a route
typedef RouteWidgetBuilder = Widget Function(BuildContext context, RouteArgs args);

/// Transform result when calling [AppRoute.openForward]
typedef ResultTransform<T, R> = FutureOr<R> Function(T value);

/// Build root theme of a route before [SystemUiOverlayStyle] is applied to the route [AnnotatedRegion]
typedef ThemeBuilder = ThemeData Function(BuildContext context, RouteSettings settings);

/// A route to a navigation UI component.
///
/// Supported routes are:
/// - [ActivityRoute]
/// - [DialogRoute]
/// - [BottomSheetRoute]
///
/// [onGenerateRoute] should be passed to [MaterialApp] onGenerateRoute parameter
/// [name] of the route
/// [open] opens this route on top of the current
/// [openReplace] replaces the current route with this one and return result for the old one
/// [openForward] replaces the current route with this one and forward result to the new one.
///               This means that old route will return the result of the new route after it's returned
///               (and maybe transformed)
/// [T] the result type for this route
abstract class AppRoute<T> {
  static const RouteFactory onGenerateRoute = ActivityRoute._onGenerateRoute;

  final String name;

  const AppRoute(this.name);

  Future<T> open(BuildContext context, {RouteArgs args = const {}});

  Future<T> openReplace(BuildContext context, {RouteArgs args = const {}, Object? result}) {
    Navigator.of(context).pop(result);
    return open(context, args: args);
  }

  Future<T> openForward<R>(BuildContext context, {RouteArgs args = const {}, ResultTransform<T, R>? transform}) {
    transform ??= (value) => value as R;
    final completer = Completer<R>();
    Navigator.of(context).pop(_ForwardResult<R>(completer.future));
    final raw = open(context, args: args);
    raw.then(transform).then(completer.complete, onError: completer.completeError);
    return raw;
  }

  static Future<T> _result<T>(Future result) async =>
      result.then((value) async => value is _ForwardResult ? await value.value : value as T);
}

class _ForwardResult<R> {
  final Future<R> value;

  const _ForwardResult(this.value);
}
